import abc
import datetime
import operator
import warnings
from collections.abc import Sequence, Mapping, Iterable
from typing import Literal, Protocol, runtime_checkable

from . import LOGGER
from ..base import MarketData, DataType, MarketDataBuffer

LOGGER = LOGGER.getChild('Replay')
__all__ = ['MarketDateCallable', 'MarketDataLoader', 'MarketDataBulkLoader', 'Replay', 'SimpleReplay', 'ProgressReplay', 'ProgressiveReplay']


@runtime_checkable
class MarketDateCallable(Protocol):
    def __call__(self, market_date: datetime.date) -> None:
        ...


@runtime_checkable
class MarketDataLoader(Protocol):
    def __call__(self, market_date: datetime.date, ticker: str, dtype: str | DataType) -> Sequence[MarketData] | Mapping[float, MarketData]:
        pass


@runtime_checkable
class MarketDataBulkLoader(Protocol):
    def __call__(self, market_date: datetime.date, tickers: Sequence[str], dtypes: Sequence[str | DataType]) -> Sequence[MarketData] | Mapping[float, MarketData] | MarketDataBuffer:
        pass


class Replay(object, metaclass=abc.ABCMeta):
    # __slots__ = ('start_date', 'end_date', 'market_date', 'calendar', 'bod', 'eod', 'subscription', '_calendar', '_market_date', '_status', '_progress')

    def __init__(self, start_date: datetime.date = None, end_date: datetime.date = None, market_date: datetime.date = None, calendar: Sequence[datetime.date] = None, bod: MarketDateCallable = None, eod: MarketDateCallable = None) -> None:
        self.start_date = start_date or market_date or calendar[0]
        self.end_date = end_date or calendar[-1]
        self.market_date = market_date or start_date
        self.calendar = calendar or []

        self.bod = []
        self.eod = []
        self.subscription = {}

        if bod is not None:
            self.add_bod(bod)

        if eod is not None:
            self.add_eod(eod)

    def add_bod(self, func: MarketDateCallable):
        self.bod.append(func)

    def add_eod(self, func: MarketDateCallable):
        self.eod.append(func)

    @classmethod
    def get_dtype(cls, dtype: DataType | str) -> str | Literal['TickData', 'TickDataLite', 'OrderData', 'TransactionData']:
        match dtype:
            case 'TickData' | 'TickDataLite' | 'OrderData' | 'TransactionData':
                return str(dtype)
            case DataType.DTYPE_TICK | DataType.DTYPE_ORDER | DataType.DTYPE_TRANSACTION:
                return DataType(dtype).name.removeprefix('DTYPE_').capitalize() + 'Data'
            case DataType.DTYPE_TICK_LITE:
                return 'Data'.join(_.capitalize() for _ in DataType(dtype).name.removeprefix('DTYPE_').split('_'))
            case _:
                raise ValueError(f'Invalid dtype {dtype}, expect str or int.')

    def add_subscription(self, ticker: str, dtype: DataType | str):
        dtype = self.get_dtype(dtype)
        topic = f'{ticker}.{dtype}'

        self.subscription[topic] = (ticker, dtype)

    def remove_subscription(self, ticker: str, dtype: DataType | str):
        dtype = self.get_dtype(dtype)
        topic = f'{ticker}.{dtype}'

        try:
            self.subscription.pop(topic)
        except KeyError as _:
            LOGGER.info(f'{topic} not in {self.subscription}')

    @abc.abstractmethod
    def __next__(self):
        ...

    @abc.abstractmethod
    def __iter__(self):
        ...


class SimpleReplay(Replay):
    def __init__(
            self,
            loader: MarketDataBulkLoader | MarketDataLoader = None,
            market_date: datetime.date = None,
            start_date: datetime.date = None,
            end_date: datetime.date = None,
            calendar: Sequence[datetime.date] = None,
            bod: MarketDateCallable = None,
            eod: MarketDateCallable = None
    ):
        super().__init__(market_date=market_date, start_date=start_date, end_date=end_date, calendar=calendar, bod=bod, eod=eod)
        self.loader = loader

    def __iter__(self):
        self._calendar = self.calendar or [self.start_date + datetime.timedelta(days=i) for i in range((self.end_date - self.start_date).days + 1)]
        self._market_date = self.market_date or sorted(_ for _ in self._calendar if _ >= self.market_date)[0]
        self._status = {market_date: 'skipped' if market_date < self.market_date else 'idle' for market_date in self._calendar}
        self._idx_buffer = 0
        self._idx_date = sum([1 for _ in self._calendar if _ < self.market_date])

        for func in self.bod:
            func(self._market_date)

        if self.loader is None:
            assert hasattr(self, '_buffer') and isinstance(self._buffer, Iterable), f'Without assigning a data loader, the _buffer of {self.__class__.__name__} should be set in bod process.'
        elif isinstance(self.loader, MarketDataLoader):
            md_list = []
            for topic, (_ticker, _dtype) in self.subscription.items():
                LOGGER.info(f'{self} loading {self._market_date} {_ticker} {_dtype}')
                data = self.loader(market_date=self._market_date, ticker=_ticker, dtype=_dtype)
                if isinstance(data, Mapping):
                    md_list.extend(list(data.values()))
                elif isinstance(data, Sequence):
                    md_list.extend(data)
                else:
                    raise TypeError(f'The loader {self.loader} returned {type(data)}. Expect a sequence or mapping of MarketData')
            md_list.sort(key=operator.attrgetter('timestamp', 'ticker', '_dtype'))
            self._buffer = iter(md_list)
            self._buffer_size = len(md_list)
        elif isinstance(self.loader, MarketDataBulkLoader):
            self._buffer = self.loader(market_date=self._market_date, tickers=self.tickers, dtypes=self.dtypes)
            self._buffer_size = len(self._buffer)
        else:
            raise NotImplementedError()

        return self

    def __next__(self) -> MarketData:
        if self._idx_buffer < self._buffer_size:
            self._idx_buffer += 1
            return next(self._buffer)

        for func in self.eod:
            func(self._market_date)

        self._idx_buffer = 0
        self._idx_date += 1

        if self._idx_date >= len(self._calendar):
            self._calendar.clear()
            del self._calendar
            del self._market_date
            del self._status
            del self._idx_buffer
            del self._idx_date
            del self._buffer
            del self._buffer_size
            raise StopIteration()

        self._market_date = self._calendar[self._idx_date]

        for func in self.bod:
            func(self._market_date)

        self._buffer = self.loader(market_date=self._market_date, tickers=self.tickers, dtypes=self.dtypes)
        return self.__next__()

    def __repr__(self):
        return f'{self.__class__.__name__}{{id={id(self)}, from={self.start_date}, to={self.end_date}}}'

    @property
    def progress(self) -> float:
        if not hasattr(self, '_buffer'):
            raise RuntimeError(f'{self.__class__.__name__} not started yet.')

        return (self._idx_date + (self._idx_buffer / self._buffer_size - 1)) / len(self._calendar)

    @property
    def tickers(self) -> list[str]:
        tickers = set()
        for _, (ticker, dtype) in self.subscription.items():
            tickers.add(ticker)
        return list(tickers)

    @property
    def dtypes(self) -> list[str]:
        dtypes = set()
        for _, (ticker, dtype) in self.subscription.items():
            dtypes.add(dtype)
        return list(dtypes)

    @property
    def status(self) -> dict[datetime.date, str]:
        if not hasattr(self, '_status'):
            raise RuntimeError(f'{self.__class__.__name__} not started yet.')

        return self._status


class ProgressReplay(SimpleReplay):
    def __init__(
            self,
            loader: MarketDataBulkLoader | MarketDataLoader = None,
            market_date: datetime.date = None,
            start_date: datetime.date = None,
            end_date: datetime.date = None,
            calendar: Sequence[datetime.date] = None,
            bod: MarketDateCallable = None,
            eod: MarketDateCallable = None,
            **tqdm_kwargs
    ):
        super().__init__(
            loader=loader,
            market_date=market_date,
            start_date=start_date,
            end_date=end_date,
            calendar=calendar,
            bod=bod,
            eod=eod
        )

        self._tqdm_kwargs = {
            'total': 1,
            'unit_scale': True,
            'unit': 'percent',
            'mininterval': 0.1,
            'miniters': 0.001,
            **tqdm_kwargs
        }
        self.add_bod(self._update_progress_bar)

    def __iter__(self):
        from tqdm.auto import tqdm
        self._pbar = tqdm(**self._tqdm_kwargs)
        iterator = super().__iter__()

        try:
            while True:
                try:
                    result = next(iterator)
                    if self._pbar:
                        self._pbar.update(self.progress)
                    yield result
                except StopIteration:
                    break
        finally:
            if self._pbar is not None:
                self._pbar.close()
                self._pbar = None

    def __next__(self) -> MarketData:
        raise RuntimeError("MarketDataBufferReplay should be used as an iterator context")

    def _update_progress_bar(self, market_date: datetime.date):
        if self._pbar:
            self._pbar.set_description(f'Replay {market_date:%Y-%m-%d} ({self._idx_date + 1} / {len(self._calendar)})')
            self._pbar.refresh()


class ProgressiveReplay(SimpleReplay):
    """
    progressively loading and replaying market data

    requires arguments
    loader: a data loading function. Expect loader = Callable(market_date: datetime.date, ticker: str, dtype: str| type) -> dict[any, MarketData]
    start_date & end_date: the given replay period
    or calendar: the given replay calendar.

    accepts kwargs:
    ticker / tickers: the given symbols to replay, expect a str| list[str]
    dtype / dtypes: the given dtype(s) of symbol to replay, expect a str | type, list[str | type]. default = all, which is (TradeData, TickData, OrderBook)
    subscription / subscribe: the given ticker-dtype pair to replay, expect a list[dict[str, str | type]]
    """

    def __init__(
            self,
            loader: MarketDataLoader,
            tickers: str | Sequence[str] = None,
            dtypes: str | DataType | Sequence[str] | Sequence[DataType] = None,
            market_date: datetime.date = None,
            start_date: datetime.date = None,
            end_date: datetime.date = None,
            calendar: Sequence[datetime.date] = None,
            bod: MarketDateCallable = None,
            eod: MarketDateCallable = None,
            **progress_config
    ) -> None:
        warnings.deprecated('User ProgressReplay instead!')
        self.loader = loader
        super().__init__(loader=loader, market_date=market_date, start_date=start_date, end_date=end_date, calendar=calendar, bod=bod, eod=eod)

        tickers = tickers or []
        dtypes = dtypes or ['TransactionData', 'TickData', 'OrderData']

        if not isinstance(loader, MarketDataLoader):
            raise TypeError('loader function has 3 requires args, market_date, ticker and dtype.')

        if isinstance(tickers, str):
            tickers = [tickers]
        elif isinstance(tickers, Iterable):
            tickers = list(tickers)
        else:
            raise TypeError(f'Invalid ticker {tickers}, expect str or list[str]')

        if isinstance(dtypes, (str, int, DataType)):
            dtypes = [dtypes]
        elif isinstance(dtypes, Iterable):
            dtypes = list(dtypes)
        else:
            raise TypeError(f'Invalid dtype {dtypes}, expect str or list[str]')

        for ticker in tickers:
            for dtype in dtypes:
                self.add_subscription(ticker=ticker, dtype=dtype)

        self.progress_config = dict(
            tasks=1,
            **progress_config
        )

    def __iter__(self):
        from ..base import Progress
        self._pbar = Progress(**self.progress_config)
        return super().__iter__()

    def __next__(self) -> MarketData:
        try:
            result = super().__next__()
            self._pbar.done_tasks = self.progress

            if (not self._pbar.tick_size) \
                    or self._pbar.progress >= self._pbar.tick_size + self._pbar.last_output \
                    or self._pbar.is_done:
                self._pbar.output()

            return result
        except StopIteration:
            if not self._pbar.is_done:
                self.progress.done_tasks = 1
                self._pbar.output()
            raise

    def _update_progress_bar(self, market_date: datetime.date):
        if self._pbar:
            self.progress.prompt = f'Replay {market_date:%Y-%m-%d} ({self._idx_date + 1} / {len(self._calendar)}):'
            self._pbar.output()
